from typing import Any, Optional, ClassVar, Callable, Awaitable
from dataclasses import dataclass
from http import HTTPStatus

from starlette.requests import Request
from starlette.responses import Response, JSONResponse
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint

@dataclass
class HTTPError(Exception):
	desc: str
	data: Optional[Any] = None

@dataclass
class BadRequest(HTTPError):
	code = 400

@dataclass
class Unauthorised(HTTPError):
	code = 401

@dataclass
class Forbidden(HTTPError):
	code = 403

@dataclass
class Conflict(HTTPError):
	code = 409

class JSONErrorMiddleware(BaseHTTPMiddleware):
	async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
		try:
			response = await call_next(request)
			return response
		except HTTPError as e:
			code: int = e.code # type: ignore
			return JSONResponse({
				'error_type': HTTPStatus(code).phrase,
				'error_description': e.desc,
				'error_data': e.data,
			}, code)



from typing import Optional, List, Union, Dict, Any, Generator
from dataclasses import dataclass, asdict

import pytest
from starlette.testclient import TestClient

from govyn import create_app

@dataclass
class Response:
	boot_value: int

class AsyncInitAPI:
	async def startup(self) -> None:
		self.required_value = 1

	async def get(self) -> Response:
		return Response(self.required_value)

@pytest.fixture
def client() -> Generator[TestClient, None, None]:
	with TestClient(create_app(AsyncInitAPI())) as c:
		yield c # type: ignore

def test_async_startup(client: TestClient) -> None:
	res = client.get('/')
	assert res.status_code == 200
	assert res.json() == asdict(Response(1))



from typing import Any, Optional, Dict, List

from starlette.applications import Starlette
from starlette.responses import JSONResponse
from starlette.routing import Route
from starlette.middleware import Middleware
from starlette.types import ASGIApp
import uvicorn

from .route_def import make_route_def
from .schemas import build_schemas
from .endpoint import make_endpoint
from .errors import JSONErrorMiddleware
from .auth import AuthBackend, AuthMiddleware

def create_app(srv: Any, name: Optional[str] = None, auth_backend: Optional[AuthBackend] = None) -> ASGIApp:
	route_defs = [ make_route_def(getattr(srv, m)) for m in dir(srv) if not m.startswith('_') and m not in { 'startup', 'shutdown' } ]

	openapi_schemas = build_schemas(route_defs, api_name = name or type(srv).__name__)
	core_routes: Any = [
		Route('/openapi/schema', lambda _: JSONResponse(openapi_schemas))
	]

	startup_func = getattr(srv, 'startup', None)
	shutdown_func = getattr(srv, 'shutdown', None)
	startup_funcs = [ startup_func ] if startup_func else []
	shutdown_funcs = [ shutdown_func ] if shutdown_func else []
	middleware: Any = [ Middleware(JSONErrorMiddleware) ]
	if auth_backend:
		middleware.append(Middleware(AuthMiddleware, auth_backend = auth_backend))

	app = Starlette(routes = core_routes + [
		Route(r.path, make_endpoint(r), methods = [ r.http_method.upper() ])
		for r in route_defs
	], on_startup = startup_funcs, on_shutdown = shutdown_funcs, middleware = middleware)

	return app

def run_app(app: ASGIApp, port: int = 80, host: str = "0.0.0.0") -> None:
	uvicorn.run(app, host = host, port = port)



from typing import Optional, List, Union, Dict, Any
from dataclasses import dataclass, asdict

import pytest
from starlette.testclient import TestClient

from govyn import create_app

@dataclass
class ScalarTypes:
	str_field: str
	int_field: int
	float_field: float
	bool_field: bool

@dataclass
class CollectionTypes:
	dict_field: Dict[str, int]
	list_field: List[bool]
	union_field: Union[str, int]

@dataclass
class OptionalTypes:
	maybe_str_field: Optional[str]

class EchoAPI:
	async def get_scalars(self, str_field: str, int_field: int, float_field: float, bool_field: bool) -> ScalarTypes:
		return ScalarTypes(str_field, int_field, float_field, bool_field)

	async def post_scalars(self, body: ScalarTypes) -> ScalarTypes:
		return body

	async def post_collections(self, body: CollectionTypes) -> CollectionTypes:
		return body

	async def get_optional(self, maybe_str_field: Optional[str]) -> OptionalTypes:
		return OptionalTypes(maybe_str_field)

	async def post_optional(self, body: OptionalTypes) -> OptionalTypes:
		return body

@pytest.fixture
def client() -> TestClient:
	return TestClient(create_app(EchoAPI()))

@pytest.fixture
def example_scalars() -> Dict[str, Any]:
	return asdict(ScalarTypes('test_str', 123, 1.5, False))

@pytest.fixture
def example_collections() -> Dict[str, Any]:
	return asdict(CollectionTypes({ 'key1': 123, 'key2': 456 }, [ True, False, True ], 'str variant'))

def test_get(client: TestClient, example_scalars: Dict[str, Any]) -> None:
	res = client.get('/scalars', params = example_scalars)
	assert res.status_code == 200
	assert res.json() == example_scalars

def test_get_missing_field(client: TestClient, example_scalars: Dict[str, Any]) -> None:
	del example_scalars['str_field']
	res = client.get('/scalars', params = example_scalars)
	assert res.status_code == 400

def test_get_invalid_types(client: TestClient, example_scalars: Dict[str, Any]) -> None:
	example_scalars['bool_field'] = 'not a bool'
	res = client.get('/scalars', params = example_scalars)
	assert res.status_code == 400

def test_post(client: TestClient, example_scalars: Dict[str, Any]) -> None:
	res = client.post('/scalars', json = example_scalars)
	assert res.status_code == 200
	assert res.json() == example_scalars

def test_post_missing_field(client: TestClient, example_scalars: Dict[str, Any]) -> None:
	del example_scalars['int_field']
	res = client.post('/scalars', json = example_scalars)
	assert res.status_code == 400

def test_post_invalid_types(client: TestClient, example_scalars: Dict[str, Any]) -> None:
	example_scalars['str_field'] = 100
	res = client.post('/scalars', json = example_scalars)
	assert res.status_code == 400

def test_post_collections(client: TestClient, example_collections: Dict[str, Any]) -> None:
	res = client.post('/collections', json = example_collections)
	assert res.status_code == 200
	assert res.json() == example_collections

def test_post_invalid_union(client: TestClient, example_collections: Dict[str, Any]) -> None:
	example_collections['union_field'] = 1.5
	res = client.post('/collections', json = example_collections)
	assert res.status_code == 400

def test_post_invalid_dict(client: TestClient, example_collections: Dict[str, Any]) -> None:
	example_collections['dict_field'] = []
	res = client.post('/collections', json = example_collections)
	assert res.status_code == 400

def test_post_invalid_list(client: TestClient, example_collections: Dict[str, Any]) -> None:
	example_collections['list_field'] = {}
	res = client.post('/collections', json = example_collections)
	assert res.status_code == 400

def test_get_optional_present(client: TestClient) -> None:
	example_optional = asdict(OptionalTypes('string'))
	res = client.get('/optional', params = example_optional)
	assert res.status_code == 200
	assert res.json() == example_optional

def test_get_optional_missing(client: TestClient) -> None:
	example_optional = asdict(OptionalTypes(None))
	res = client.get('/optional', params = example_optional)
	assert res.status_code == 200
	assert res.json() == example_optional

def test_post_optional_present(client: TestClient) -> None:
	example_optional = asdict(OptionalTypes('string'))
	res = client.post('/optional', json = example_optional)
	assert res.status_code == 200
	assert res.json() == example_optional

def test_post_optional_missing(client: TestClient) -> None:
	example_optional = asdict(OptionalTypes(None))
	res = client.post('/optional', json = example_optional)
	assert res.status_code == 200
	assert res.json() == example_optional



from typing import Optional, List, Union, Dict
from dataclasses import dataclass, asdict

import pytest
from starlette.testclient import TestClient

from govyn import create_app
from govyn.auth import Principal, HeaderAuthBackend

class HardcodedAuthBackend(HeaderAuthBackend):
	def __init__(self) -> None:
		super().__init__("Govyn-Token")
		self.tokens = {
			'1234': 'user1',
			'5678': 'user2',
		}

	async def principal_from_header(self, token: str) -> Optional[Principal]:
		principal_id = self.tokens.get(token)
		if principal_id is None:
			return None

		return Principal(principal_id)

@dataclass
class AuthedResponse:
	name: str

@dataclass
class AuthAPI:
	async def get(self, principal: Principal) -> AuthedResponse:
		return AuthedResponse(principal.id)

@pytest.fixture
def client() -> TestClient:
	return TestClient(create_app(AuthAPI(), auth_backend = HardcodedAuthBackend()))

def test_token(client: TestClient) -> None:
	res = client.get('/', headers = { 'Govyn-Token': '1234' })
	assert res.status_code == 200
	assert res.json() == asdict(AuthedResponse('user1'))
	
	res = client.get('/', headers = { 'Govyn-Token': '5678' })
	assert res.status_code == 200
	assert res.json() == asdict(AuthedResponse('user2'))

def test_token_invalid(client: TestClient) -> None:
	res = client.get('/', headers = { 'Govyn-Token': 'not a token' })
	assert res.status_code == 401

def test_token_missing(client: TestClient) -> None:
	res = client.get('/')
	assert res.status_code == 401



from typing import Any, Union, Dict, Callable
from dataclasses import dataclass

from .errors import BadRequest
from .auth import Principal

_ParserType = Callable[[ str ], Any]

@dataclass
class ArgDef:
	expected_type: type
	optional: bool
	parser: _ParserType

def _parse_bool(x: str) -> bool:
	x = x.lower()
	if x == "true":
		return True
	elif x == "false":
		return False
	raise BadRequest('invalid boolean value')

_parser_overrides: Dict[type, _ParserType] = {
	bool: _parse_bool,
}

def make_arg_def(expected_type: type) -> ArgDef:
	is_optional = getattr(expected_type, '__origin__', None) == Union and getattr(expected_type, '__args__')[1] == type(None)

	if is_optional:
		expected_type = getattr(expected_type, '__args__')[0]

	parser = _parser_overrides.get(expected_type, expected_type)

	return ArgDef(
		expected_type = expected_type,
		optional = is_optional,
		parser = parser,
	)

@dataclass
class RouteDef:
	path: str
	http_method: str
	impl: Callable[..., Any]
	args: Dict[str, ArgDef]
	return_type: type
	requires_principal: bool

def make_route_def(impl: Callable[..., Any]) -> RouteDef:
	name_tokens = impl.__name__.split('_')
	http_method = name_tokens[0]

	input_annotations = impl.__annotations__.copy()
	return_type = input_annotations['return']
	del input_annotations['return']
	requires_principal = input_annotations.get('principal') is not None
	if requires_principal:
		del input_annotations['principal']

	if http_method == 'post':
		if len(input_annotations) != 1:
			raise Exception('POST methods require one argument')

	args = {
		name: make_arg_def(expected_type) for name, expected_type
		in input_annotations.items()
	}

	return RouteDef(
		path = '/' + '_'.join(name_tokens[1:]),
		http_method = http_method,
		impl = impl,
		args = args,
		return_type = return_type,
		requires_principal = requires_principal,
	)



from setuptools import setup, find_packages

setup(
	name = 'govyn',
	version = '0.1.0',
	author = 'Ruan Pearce-Authers',
	author_email = 'ruanpa@outlook.com',
	description = 'HTTP APIs in typed Python',
	url = 'https://github.com/returnString/govyn',
	packages = find_packages(),
	package_data = {
		'govyn': [ 'py.typed' ],
	},
	zip_safe = False,
	install_requires = [
		'starlette >= 0.13',
		'uvicorn >= 0.11',
		'dacite >= 1.5.1',
		'requests >= 2.24',
	],
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
	python_requires = '>= 3.8',
)



from typing import Set, Optional, Dict, Protocol
from dataclasses import dataclass
from abc import ABC, abstractmethod

from starlette.types import ASGIApp
from starlette.requests import Request
from starlette.responses import Response
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint

from .errors import Unauthorised

@dataclass
class Principal:
	id: str

class AuthBackend(Protocol):
	async def resolve_principal(self, req: Request) -> Optional[Principal]:
		...

class AuthMiddleware(BaseHTTPMiddleware):
	def __init__(self, app: ASGIApp, auth_backend: AuthBackend) -> None:
		super().__init__(app)
		self.auth_backend = auth_backend

	async def dispatch(self, req: Request, call_next: RequestResponseEndpoint) -> Response:
		principal = await self.auth_backend.resolve_principal(req)
		if principal is None:
			raise Unauthorised('authentication failed')

		req.state.principal = principal
		return await call_next(req)

class HeaderAuthBackend(ABC):
	def __init__(self, header: str) -> None:
		self.header = header

	async def resolve_principal(self, req: Request) -> Optional[Principal]:
		token = req.headers.get(self.header)
		if not token:
			raise Unauthorised(f'missing header: {self.header}')

		return await self.principal_from_header(token)

	@abstractmethod
	async def principal_from_header(self, value: str) -> Optional[Principal]:
		...



from typing import Dict, Any
from dataclasses import asdict

from starlette.requests import Request
from starlette.responses import JSONResponse
from dacite.core import from_dict
from dacite.exceptions import DaciteError

from .route_def import RouteDef, ArgDef
from .errors import BadRequest

async def query_string_parser(req: Request, args: Dict[str, ArgDef]) -> Dict[str, Any]:
	ret: Dict[str, Any] = dict()
	for var_name, arg_def in args.items():
		value = req.query_params.get(var_name)
		if value is None:
			if not arg_def.optional:
				raise BadRequest(f'missing required field: {var_name}')
			ret[var_name] = None
		else:
			try:
				ret[var_name] = arg_def.parser(value) # type: ignore
			except ValueError as e:
				raise BadRequest(str(e))

	return ret

async def json_body_parser(req: Request, args: Dict[str, ArgDef]) -> Dict[str, Any]:
	json_body = await req.json()
	name = list(args)[0]
	arg_def = args[name]

	try:
		body: Any = from_dict(arg_def.expected_type, json_body)
	except DaciteError as e:
		raise BadRequest(str(e))

	return { name: body }

_parser_dict = {
	'get': query_string_parser,
	'post': json_body_parser,
}

def make_endpoint(route: RouteDef) -> Any:
	parser = _parser_dict[route.http_method]

	async def endpoint(req: Request) -> Any:
		args = await parser(req, route.args)
		if route.requires_principal:
			args['principal'] = req.state.principal
		res = await route.impl(**args)
		return JSONResponse(asdict(res))

	return endpoint
